;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; toolsaq.lisp

;;;; This file contains the main (exported) queries available when using
;;;; the ToolSAQ (TOOL for Spliced Alignment Queries) environment.

;;;; Michael E Sparks (mespar1@iastate.edu)
;;;; Last Modified: 3 February 2007

;;;; Copyright (c) 2007 Michael E Sparks
;;;; All rights reserved.
;;;;
;;;; You are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser General Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.

(in-package :toolsaq)

(defun show-toolsaq-exports (&optional (stream *standard-output*))
  "Prints symbols exported from ToolSAQ to the specified stream."
  (do-external-symbols (sym :toolsaq)
    (format stream "~&")
    (describe sym stream)
    (format stream "~%~%")))

(defconstant *default-gthxml-file-extension* "xml"
  "Consider any file with this extension to be a valid gthXML formatted document.")

(defun operate-on-container (operation out-file-extension
			     &optional (gthxml-container
					nil
					gthxml-container-supplied-p))
  "Applies operation to gthXML documents in a gthxml-container, which
are identified by having a file extension of *default-gthxml-file-extension*;
operation should return an object with a print-object method specialized
on it.  Output will be written to a file with the same name as the
gthXML input file, but with an added extension of out-file-extension.
Because this function works primarily via side effects, its return value
is undefined.

Note that it is assumed the gthXML documents have previously been externally
validated against the gthXML v.1.1 grammar."
  (flet ((get-document-container ()
	   (pathname-directory
	    (parse-namestring
	     (if gthxml-container-supplied-p
		 gthxml-container
		 (progn (format *query-io*
				"~&Please supply location of gthXML file container: ")
			(force-output *query-io*)
			;; On my system, I get an auto-complete if I enter input strings
			;; starting with #\", which is rather handy, IMHO, but introduces
			;; contaminants.  If #\" is a character in the pathname,
			;; the following expression would have to be re-written.
			(remove #\" (read-line *query-io*))))))))
    (dolist (current-file
	      (directory
	       (make-pathname
		:directory (get-document-container)
		:name :wild :type *default-gthxml-file-extension*)))
      (with-open-file (gthxml-stream current-file)
	(with-open-file
	    (results-stream
	     (make-pathname :name (concatenate 'string
					       (pathname-name current-file) "."
					       (pathname-type current-file))
			    :type out-file-extension
			    :defaults current-file)
	     :direction :output :if-exists :supersede)
	  (princ (funcall operation gthxml-stream) results-stream))))))

(defmacro detect-form (fct-to-apply extension container container-p)
  "This macro wraps the general form of detection methods available from ToolSAQ."
  `(let ((fct ,fct-to-apply)
	 (ext ,extension)
	 (cont ,container))
    (declare (ignorable cont))
    (check-type fct function)
    (check-type ext string)
    (let ((arg-list (list fct ext)))
      (apply #'operate-on-container
	     (if ,container-p
		 (progn (check-type cont string)
			(append arg-list (list cont)))
		 arg-list)))))

;;; Alternative splicing operations (based on AGS data)

(defconstant *ags-based-alt-splc-extension* "altsplc"
  "Default filename extension for result files of AGS-based alternative splice detection.")

(defun detect-alternative-splicing (&key (gthxml-container nil gthxml-container-supplied-p))
  "Identifies alternatively spliced transcript isoforms at eukaryotic
genetic loci by considering alternative gene structures, deduced using
the maximal transcript alignment assembly algorithm implemented in, e.g.,
GenomeThreader.  The parameter *ags-alt-splc-parms* is of particular
relevance to this process."
  (validate-parms-obj-qual *qual-aln-parms*)
  (validate-parms-obj-qual *ags-alt-splc-parms*)
  (detect-form #'find-altsplc-from-ags
	       *ags-based-alt-splc-extension*
	       gthxml-container
	       gthxml-container-supplied-p))

;;; Operations for Alternative Gene Structures (AGSs) in general

(defconstant *all-ags-extension* "allags"
  "Default filename extension for result files of parsing
all AGSs in a gthXML document.")

(defun detect-all-agss (&key (gthxml-container nil gthxml-container-supplied-p))
  "Identifies all alternative gene structures in a gthXML document."
  (detect-form #'find-all-ags
	       *all-ags-extension*
	       gthxml-container
	       gthxml-container-supplied-p))

;;; Operations for parsing query (transcript or protein)
;;; alignments (PGS and (optionally) hqPGS) from gthXML
;;; documents.

(defmacro aln-closure-factory (fct)
  "For use with the alignment detectors, this macro
expands into code for a closure that takes an input
stream, applies fct to it, and captures its output
to a string, which it returns."
  `#'(lambda (stream)
       (with-output-to-string (accumulator)
	 (funcall (,fct stream)
		  accumulator))))

;;; For finding only "qualified" alignments

(defconstant *qualified-alns-extension* "qualalns"
  "Default filename extension for result files of parsing
qualifed PGSs, and possibly hqPGSs, from a gthXML document.")

(defun detect-qualified-alns (&key (gthxml-container nil gthxml-container-supplied-p))
  "Identifies all query, i.e., transcript or protein sequence, alignments
in a gthXML document, being PGS and possibly hqPGS data, subject to
constraints specified in the *qual-aln-parms* parameter object."
  (validate-parms-obj-qual *qual-aln-parms*)
  (detect-form (aln-closure-factory parse-qualified-alns-from-gthxml)
	       *qualified-alns-extension*
	       gthxml-container
	       gthxml-container-supplied-p))

;;; For finding all alignments

(defconstant *all-alns-extension* "allalns"
  "Default filename extension for result files of parsing
all PGSs, and possibly hqPGSs, from a gthXML document.")

(defun detect-all-alns (&key (gthxml-container nil gthxml-container-supplied-p))
  "Identifies all query, i.e., transcript or protein sequence, alignments
in a gthXML document, being PGS and possibly hqPGS data."
  (detect-form (aln-closure-factory parse-all-alns-from-gthxml)
	       *all-alns-extension*
	       gthxml-container
	       gthxml-container-supplied-p))
